@@ -234,6 +234,27 @@ def Batch(batch_size):  # pylint: disable=invalid-name
   return lambda g: batch(g, batch_size)
 
 
+@gin.configurable(module='trax.data')
+def Dup():  # pylint: disable=invalid-name
+  """Duplicates (copies) the top element (inputs).
+
+  The generator stream is augmented in the following way:
+
+  - If the stream consists of a single element `(inputs, )`,
+    the inputs simply get copied to `(inputs, inputs)`.
+  - If the stream consists of multiple elements, for example
+    `(inputs, weights)`, the rest of elements get moved toward
+    the right side `(inputs, inputs, weights)`.
+
+  Returns:
+    the duplicating function.
+  """
+  def _copy(xs):
+    x, *rest = xs
+    return (x, x, *rest)
+  return lambda g: map(lambda x: _copy(x), g)  # pylint: disable=unnecessary-lambda
+
+
 @gin.configurable(module='trax.data')
 def FilterEmptyExamples(axes=None, debug=False):  # pylint: disable=invalid-name
   """Filters empty examples.
@@ -370,7 +391,8 @@ def PadToLength(  # pylint: disable=invalid-name
           raise ValueError(f'example isn\'t nparray, but should be: {example}')
         array_length = example.shape[0]
         if multiple:
-          padding_len = array_length - ((array_length // value) * value)
+          padding_len = (
+              array_length - ((array_length // len_map[0]) * len_map[0]))
         else:
           padding_len = max(0, len_map[0] - array_length)
         output = np.pad(example,
@@ -587,7 +609,7 @@ def UniformlySeek(name=None, host_id=None, n_hosts=None, dataset_size=None):  #
     logging.error(
         'No dataset size given to Uniformly seek, assuming: %d', dataset_size)
   assert name
-  host_id = jax.host_id() if host_id is None else host_id
+  host_id = jax.process_index() if host_id is None else host_id
   n_hosts = n_hosts or jax.host_count()
   each_host = int(dataset_size / n_hosts)
   def _f(generator):
@@ -898,11 +920,12 @@ def generate_sequential_chunks(max_length=None):
       n_tokens = len(example)
       if n_tokens <= max_length:
         yield example
-      n_segments = int(math.ceil(float(n_tokens) / float(max_length)))
-      for i in range(n_segments):
-        start = max_length * i
-        end = min(start + max_length, n_tokens)
-        yield example[start:end]
+      else:
+        n_segments = int(math.ceil(float(n_tokens) / float(max_length)))
+        for i in range(n_segments):
+          start = max_length * i
+          end = min(start + max_length, n_tokens)
+          yield example[start:end]
   return _f
 
 
@@ -1073,7 +1096,7 @@ def count_and_skip(generator, name):
 def save_data_counters(output_dir, host_id=None):
   """Checkpoint data counters."""
   global data_counters
-  host_id = jax.host_id() if host_id is None else host_id
+  host_id = jax.process_index() if host_id is None else host_id
   fname = os.path.join(output_dir, 'data_counters%d.pkl' % host_id)
   with tf.io.gfile.GFile(fname, 'wb') as f:
     pickle.dump(data_counters, f)
@@ -1082,7 +1105,7 @@ def save_data_counters(output_dir, host_id=None):
 def load_data_counters(output_dir, host_id=None):
   """Checkpoint data counters."""
   global data_counters
-  host_id = jax.host_id() if host_id is None else host_id
+  host_id = jax.process_index() if host_id is None else host_id
   fname = os.path.join(output_dir, 'data_counters%d.pkl' % host_id)
   if not tf.io.gfile.exists(fname):
     logging.info('Did not load data counters as %s does not exist.', fname)
