@@ -1,5 +1,5 @@
 # coding=utf-8
-# Copyright 2020 The Trax Authors.
+# Copyright 2021 The Trax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -179,19 +179,25 @@ JAX_BACKEND = {
     'custom_vjp': _custom_vjp,
     'custom_grad': _custom_grad,
     'dataset_as_numpy': _dataset_as_numpy,
-    'device_count': jax.local_device_count,
+    'dynamic_slice': jax.lax.dynamic_slice,
+    'dynamic_slice_in_dim': jax.lax.dynamic_slice_in_dim,
+    'dynamic_update_slice': jax.lax.dynamic_update_slice,
+    'dynamic_update_slice_in_dim': jax.lax.dynamic_update_slice_in_dim,
     'erf': jax_special.erf,
     'expit': jax_special.expit,
     'fori_loop': lax.fori_loop,
+    'global_device_count': jax.device_count,
     'grad': jax.grad,
     'value_and_grad': jax.value_and_grad,
-    'index_add': jax.ops.index_add,
-    'index_max': jax.ops.index_max,
-    'index_min': jax.ops.index_min,
-    'index_update': jax.ops.index_update,
+    'index_add': lambda x, idx, y: jnp.asarray(x).at[idx].add(y),
+    'index_max': lambda x, idx, y: jnp.asarray(x).at[idx].max(y),
+    'index_min': lambda x, idx, y: jnp.asarray(x).at[idx].min(y),
+    'index_update': lambda x, idx, y: jnp.asarray(x).at[idx].set(y),
     'jit': jax.jit,
+    'local_device_count': jax.local_device_count,
     'logsumexp': jax_special.logsumexp,
     'lt': lax.lt,
+    'map': lax.map,
     'max_pool': jax_max_pool,
     'pmap': jax.pmap,
     'psum': lax.psum,
@@ -207,6 +213,7 @@ JAX_BACKEND = {
     'sort_key_val': jax.lax.sort_key_val,
     'stop_gradient': lax.stop_gradient,
     'sum_pool': jax_sum_pool,
+    'top_k': lax.top_k,
     'vjp': jax.vjp,
     'vmap': jax.vmap,
 }
