@@ -1,5 +1,5 @@
 # coding=utf-8
-# Copyright 2020 The Trax Authors.
+# Copyright 2021 The Trax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -14,10 +14,11 @@
 # limitations under the License.
 
 # Lint as: python3
-"""Core layer types, such as `Dense`, `Embedding`, and `Dropout`."""
+"""Core layer types and key functions used by various layers."""
 
 from absl import logging
 import numpy as np
+import tensorflow as tf
 
 from trax import fastmath
 from trax.fastmath import numpy as jnp
@@ -49,7 +50,8 @@ class Dense(base.Layer):
                n_units,
                kernel_initializer=init.GlorotUniformInitializer(),
                bias_initializer=init.RandomNormalInitializer(1e-6),
-               use_bias=True):
+               use_bias=True,
+               use_bfloat16=False):
     """Returns a dense (fully connected) layer of width `n_units`.
 
     A dense layer maps collections of `R^m` vectors to `R^n`, where `n`
@@ -65,12 +67,15 @@ class Dense(base.Layer):
           bias weights `b` for the layer.
       use_bias: If `True`, compute an affine map `y = Wx + b`; else compute
           a linear map `y = Wx`.
+      use_bfloat16: If `True`, use bfloat16 weights instead of the default
+        float32; this can save memory but may (rarely) lead to numerical issues.
     """
     super().__init__(name=f'Dense_{n_units}')
     self._n_units = n_units
     self._kernel_initializer = kernel_initializer
     self._bias_initializer = bias_initializer
     self._use_bias = use_bias
+    self._use_bfloat16 = use_bfloat16
 
   def forward(self, x):
     """Executes this layer as part of a forward pass through the model.
@@ -107,9 +112,13 @@ class Dense(base.Layer):
     shape_b = (self._n_units,)
     rng_w, rng_b = fastmath.random.split(self.rng, 2)
     w = self._kernel_initializer(shape_w, rng_w)
+    if self._use_bfloat16:
+      w = w.astype(jnp.bfloat16)
 
     if self._use_bias:
       b = self._bias_initializer(shape_b, rng_b)
+      if self._use_bfloat16:
+        b = b.astype(jnp.bfloat16)
       self.weights = (w, b)
     else:
       self.weights = w
@@ -119,7 +128,7 @@ class Dense(base.Layer):
 # dimension at the end. This dimension size corresponds to embedding depth.
 @assert_shape('...->...d')
 class Embedding(base.Layer):
-  """Trainable layer that maps discrete tokens/ids to vectors.
+  """Trainable layer that maps discrete tokens/IDs to vectors.
 
   Embedding layers are commonly used to map discrete data, like words in NLP,
   into vectors. Here is a canonical example::
@@ -134,42 +143,46 @@ class Embedding(base.Layer):
   def __init__(self,
                vocab_size,
                d_feature,
-               kernel_initializer=
-               init.ScaledInitializer(out_dim=-1,
-                                      in_dim=-2,
-                                      scale=1.,
-                                      mode='fan_out',
-                                      distribution='uniform')):
+               use_bfloat16=False,
+               kernel_initializer=init.ScaledInitializer(
+                   out_dim=-1, in_dim=-2, scale=1., mode='fan_out',
+                   distribution='uniform')):
     """Returns an embedding layer with given vocabulary size and vector size.
 
-    The layer clips input values (token ids) to the range `[0, vocab_size)`.
-    That is, negative token ids all clip to `0` before being mapped to a
-    vector, and token ids with value `vocab_size` or greater all clip to
+    The layer clips input values (token IDs) to the range `[0, vocab_size)`.
+    That is, negative token IDs all clip to `0` before being mapped to a
+    vector, and token IDs with value `vocab_size` or greater all clip to
     `vocab_size - 1` before being mapped to a vector.
 
     Args:
       vocab_size: Size of the input vocabulary. The layer will assign a unique
-          vector to each id in `range(vocab_size)`.
+        vector to each id in `range(vocab_size)`.
       d_feature: Dimensionality/depth of the output vectors.
+      use_bfloat16: If `True`, use bfloat16 weights instead of the default
+        float32; this can save memory but may (rarely) lead to numerical issues.
       kernel_initializer: Function that creates (random) initial vectors for
-          the embedding.
+        the embedding.
     """
     # TODO(jonni): is the clipping behavior what we want going forward?
     super().__init__(name=f'Embedding_{vocab_size}_{d_feature}')
     self._d_feature = d_feature  # feature dimensionality
     self._vocab_size = vocab_size
+    self._use_bfloat16 = use_bfloat16
     self._kernel_initializer = kernel_initializer
 
   def forward(self, x):
-    """Returns embedding vectors corresponding to input token id's.
+    """Returns embedding vectors corresponding to input token IDs.
 
     Args:
-      x: Tensor of token id's.
+      x: Tensor of token IDs.
 
     Returns:
       Tensor of embedding vectors.
     """
-    return jnp.take(self.weights, x, axis=0)
+    embedded = jnp.take(self.weights, x, axis=0, mode='clip')
+    if self._use_bfloat16:  # Return float32 activations w/ bfloat16 weights.
+      embedded = embedded.astype(jnp.float32)
+    return embedded
 
   def init_weights_and_state(self, input_signature):
     """Randomly initializes this layer's weights."""
@@ -177,6 +190,8 @@ class Embedding(base.Layer):
     shape_w = (self._vocab_size, self._d_feature)
     # TODO(lukaszkaiser): do we split self.rng for consistency? Add a method?
     w = self._kernel_initializer(shape_w, self.rng)
+    if self._use_bfloat16:
+      w = w.astype(jnp.bfloat16)
     self.weights = w
 
 
@@ -234,6 +249,8 @@ class Dropout(base.Layer):
     rate = self._initial_rate
     if isinstance(state, dict) and self._name in state:
       rate = state[self._name]
+    if rate == 0.0:
+      return x
     mask_shape = list(x.shape)
     for axis in self._shared_axes:
       mask_shape[axis] = 1
@@ -249,16 +266,19 @@ class Weights(base.Layer):
   It takes no input and returns a single tensor: weights.
   """
 
-  def __init__(self, initializer, shape=tuple()):
+  def __init__(self, initializer, shape=tuple(), use_bfloat16=False):
     """Returns a learnable tensor of shape `shape`.
 
     Args:
       initializer: Function taking shape and rng as arguments.
       shape: Shape of the learnable weights.
+      use_bfloat16: If `True`, use bfloat16 weights instead of the default
+        float32; this can save memory but may (rarely) lead to numerical issues.
     """
     super().__init__(name=f'Weights_{shape}', n_in=0, n_out=1)
     self._shape = shape
     self._initializer = initializer
+    self._use_bfloat16 = use_bfloat16
 
   def forward(self, x):
     """Executes this layer as part of a forward pass through the model.
@@ -284,12 +304,19 @@ class Weights(base.Layer):
     """
     del input_signature  # Unused. There is no input to this layer.
     self.weights = self._initializer(self._shape, self.rng)
+    if self._use_bfloat16:
+      self.weights = self.weights.astype(jnp.bfloat16)
 
 
 def PrintShape(n_in=1, msg=''):
   """Prints the shapes of `n_in` inputs and returns then unchanged."""
   def Fwd(xs):
-    shapes_and_dtypes = ', '.join([str(x.shape) + f'[{x.dtype}]' for x in xs])
+    def format_shape(x):  # pylint: disable = invalid-name
+      return str(x.shape) + f'[{x.dtype}]'
+    if n_in > 1:
+      shapes_and_dtypes = ', '.join([format_shape(x) for x in xs])
+    else:
+      shapes_and_dtypes = format_shape(xs)
     info = f'PrintShape: {msg}: [{shapes_and_dtypes}]'
     print(info)
     logging.info(info)
@@ -297,6 +324,67 @@ def PrintShape(n_in=1, msg=''):
   return base.PureLayer(Fwd, n_in=n_in, n_out=n_in, name=f'PrintShape_{n_in}')
 
 
+class SummaryImage(base.Layer):
+  """A layer receiving a tensor, and adding it to TensorBoard as an image.
+
+  It takes an input and returns it unchanged. It stores this input as a state to
+  be used as a metric in TensorBoard.
+  It converts a tensor to a scalar by running a given aggregation function (mean
+  by default). On TensorBoard, results for each device will be reported
+  separately.
+  """
+
+  def __init__(self, name, n_in, num_summaries=5,
+               recover_fn=None):
+    """Takes a tensor and returns it.
+
+    Args:
+      name: Name of the metric to be reported.
+      n_in: Number of inputs.
+      num_summaries: Number of images to show.
+      recover_fn: the function for converting a tensor to a dipslayable image.
+    """
+    super().__init__(name=f'Summary_{name}', n_in=n_in, n_out=n_in)
+    name = 'summary_' + name
+    self._name = name
+    self._num_summaries = num_summaries
+    self._recover_fn = recover_fn
+
+  def forward(self, x):
+    """Executes this layer as part of a forward pass through the model.
+
+    Args:
+      x: Tensor of same shape and dtype as the input signature used to
+          initialize this layer.
+
+    Returns:
+      Tensor with previously specified shape and dtype.
+    """
+    self.state = {}
+    batch_size = x[0].shape[0]
+    num_images = min(self._num_summaries, batch_size)
+    for s in range(num_images):
+      images = []
+      for i in range(self._n_in):
+        images.append(
+            self._recover_fn(x[i][s]) if self._recover_fn else x[i][s])
+      self.state[self._name + str(s)] = jnp.concatenate(images, axis=0)
+    return x[:self._n_in]
+
+  def init_weights_and_state(self, input_signature):
+    """Returns newly initialized weights for this layer.
+
+    Weights is a single  `w` tensor with previously specified shape.
+
+    Args:
+      input_signature: `ShapeDtype` instance characterizing the input this layer
+          should compute on. Unused.
+    """
+    del input_signature  # Unused.
+    self.weights = ()
+    self.state = {self._name: jnp.array(0.)}
+
+
 class SummaryScalar(base.Layer):
   """A layer receiving a tensor, and adding it to TensorBoard as a scalar.
 
@@ -382,7 +470,7 @@ class RandomUniform(base.Layer):
     return result
 
   def _get_conditionally_synced_rng(self):
-    if self._sync and fastmath.device_count() > 1:
+    if self._sync and fastmath.global_device_count() > 1:
       return fastmath.psum(self.rng, 'batch')
     else:
       return self.rng
@@ -551,19 +639,19 @@ def Flatten(n_axes_to_keep=1):
     if in_rank <= n_axes_to_keep:
       raise ValueError(f'Input rank ({in_rank}) must exceed the number of '
                        f'axes to keep ({n_axes_to_keep}) after flattening.')
-    return jnp.reshape(x, (x.shape[:n_axes_to_keep] + (-1,)))
+    shape = x.shape
+    if isinstance(shape, tf.TensorShape):
+      shape = tuple(shape.as_list())
+    return jnp.reshape(x, (shape[:n_axes_to_keep] + (-1,)))
   return Fn(layer_name, f)
 
 
-@assert_shape('...->...')  # The output and input shapes are the same.
-def Exp():
-  """Returns a layer that computes the element-wise exponential of a tensor."""
-  return Fn('Exp', lambda x: jnp.exp(x))  # pylint: disable=unnecessary-lambda
-
-
 def LogSoftmax(axis=-1):
   """Returns a layer that applies log softmax along one tensor axis.
 
+  Note that the implementation actually computes x - LogSumExp(x),
+  which is mathematically equal to LogSoftmax(x).
+
   `LogSoftmax` acts on a group of values and normalizes them to look like a set
   of log probability values. (Probability values must be non-negative, and as
   a set must sum to 1. A group of log probability values can be seen as the
@@ -572,8 +660,17 @@ def LogSoftmax(axis=-1):
   Args:
     axis: Axis along which values are grouped for computing log softmax.
   """
-  return Fn('LogSoftmax',
-            lambda x: x - fastmath.logsumexp(x, axis, keepdims=True))
+  return Fn('LogSoftmax', lambda x: log_softmax(x, axis=axis))
+
+
+def LogSumExp(axis=-1):
+  """Returns a layer that computes log(sum(exp(x))) along one tensor axis.
+
+  Args:
+    axis: Axis along which values are grouped for computing log-sum-exp.
+  """
+  return Fn('LogSumExp',
+            lambda x: fastmath.logsumexp(x, axis=axis, keepdims=True))
 
 
 def Softmax(axis=-1):
@@ -587,7 +684,7 @@ def Softmax(axis=-1):
     axis: Axis along which values are grouped for computing softmax.
   """
   return Fn('Softmax',
-            lambda x: jnp.exp(x - fastmath.logsumexp(x, axis, keepdims=True)))
+            lambda x: jnp.exp(log_softmax(x, axis=axis)))
 
 
 def ToFloat():
@@ -634,7 +731,7 @@ def Max(axis=-1, keepdims=False):
   return Fn('Max', lambda x: jnp.max(x, axis, keepdims=keepdims))
 
 
-def Sum(axis=-1, keepdims=False):
+def Sum(axis=None, keepdims=False):
   """Returns a layer that computes sums using one tensor axis.
 
   `Sum` uses one tensor axis to form groups of values and replaces each group
@@ -644,13 +741,28 @@ def Sum(axis=-1, keepdims=False):
   one.
 
   Args:
-    axis: Axis along which values are grouped for computing a sum.
+    axis: Axis along which values are grouped for computing a sum; if None,
+        compute sum over all elements in tensor.
     keepdims: If `True`, keep the resulting size 1 axis as a separate tensor
         axis; else, remove that axis.
   """
   return Fn('Sum', lambda x: jnp.sum(x, axis=axis, keepdims=keepdims))
 
 
+def ThresholdToBinary(threshold=.5):
+  """Returns a layer that thresholds inputs to yield outputs in {0, 1}."""
+  def f(model_output):  # pylint: disable=invalid-name
+    return (model_output > threshold).astype(jnp.int32)
+  return Fn('ThresholdToBinary', f)
+
+
+def ArgMax(axis=-1):
+  """Returns a layer that calculates argmax along the given axis."""
+  def f(model_output):  # pylint: disable=invalid-name
+    return jnp.argmax(model_output, axis=axis)
+  return Fn('ArgMax', f)
+
+
 @assert_shape('...->...')  # The output and input shapes are the same.
 def Negate():
   """Returns a layer that computes the element-wise negation of a tensor."""
@@ -663,6 +775,30 @@ def StopGradient():
   return Fn('StopGradient', lambda x: fastmath.stop_gradient(x))  # pylint: disable=unnecessary-lambda
 
 
+def one_hot(x, n_categories, dtype=jnp.float32):  # pylint: disable=invalid-name
+  """Makes a one-hot array (n+1 dims) from an int-categorical array (n dims)."""
+  indices_less_than_n = jnp.arange(n_categories)
+  return jnp.array(x[..., jnp.newaxis] == indices_less_than_n, dtype)
+
+
+def log_softmax(x, axis=-1):  # pylint: disable=invalid-name
+  """Transforms activation vectors to log-probability vectors.
+
+  Log probability vectors are derived by, in effect, applying softmax to raw
+  activation vectors and then applying log element-wise. The actual
+  implementation uses a mathematically valid simplification of this.
+
+  Args:
+    x: An ndarray with activation vectors along the given axis.
+    axis: Axis along which values are grouped for computing log softmax.
+
+  Returns:
+    An ndarray containing log-probability vectors derived from the raw
+    activation vectors in `x`.
+  """
+  return x - fastmath.logsumexp(x, axis=axis, keepdims=True)
+
+
 def log_gaussian_pdf(x, mu, sigma):  # pylint: disable=invalid-name
   """Returns `log N(x | mu, sigma)`.
 
@@ -720,6 +856,7 @@ def multigaussian_loss(preds, targets, ngauss=1):  # pylint: disable=invalid-nam
   return fastmath.logsumexp(loglogits + glogprobs, axis=-1)
 
 
+# TODO(jonni): Rename to log_softmax_sample.
 def logsoftmax_sample(log_probs, temperature=1.0):  # pylint: disable=invalid-name
   """Returns a sample from a log-softmax output, with temperature.
 
