@@ -660,6 +660,8 @@ def train(output_dir,
         additional_eval_tasks + additional_eval_tasks_from_streams,
         output_dir=output_dir,
         checkpoint_at=checkpoint_at,
+        checkpoint_low_metric=checkpoint_lowest,
+        checkpoint_high_metric=checkpoint_highest,
         permanent_checkpoint_at=permanent_checkpoint_at,
         n_devices=n_devices,
         loss_chunk_size=loss_chunk_size,
