@@ -20,9 +20,9 @@ import os
 
 from absl.testing import absltest
 from absl.testing import parameterized
+import jax
 from jax import test_util  # pylint: disable=unused-import
 from jax.config import config
-from jax.lib import xla_bridge
 import tensorflow.compat.v2 as tf
 from trax import fastmath
 from trax import layers as tl
@@ -41,7 +41,7 @@ from trax.tf_numpy import numpy as tf_np
 
 def _test_inputs(n_classes, with_weights=False, input_shape=(6, 6, 3)):
   """Make trainer_lib.inputs.Inputs."""
-  batch_size = 2 * xla_bridge.device_count()
+  batch_size = 2 * jax.device_count()
 
   def input_stream(n_devices):
     del n_devices
@@ -67,7 +67,7 @@ def _test_inputs(n_classes, with_weights=False, input_shape=(6, 6, 3)):
 
 def _test_inputs_lm(vocab_size, seq_len, per_device_batch_size=2):
   """Make trainer_lib.inputs.Inputs for language model."""
-  batch_size = per_device_batch_size * xla_bridge.device_count()
+  batch_size = per_device_batch_size * jax.device_count()
 
   def input_stream(_):
     def make_batch(key):
