@@ -1,5 +1,5 @@
 # coding=utf-8
-# Copyright 2020 The Trax Authors.
+# Copyright 2021 The Trax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -14,12 +14,9 @@
 # limitations under the License.
 
 """Trax trainer."""
-
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import print_function
-
+import atexit
 import datetime
+import functools
 import os
 
 from absl import app
@@ -28,6 +25,7 @@ from absl import logging
 
 import gin
 import jax
+from jax.lib import xla_extension as xc
 import tensorflow.compat.v2 as tf
 from trax import fastmath
 from trax import trainer_flags  # pylint: disable=unused-import
@@ -45,7 +43,7 @@ def _tf_setup_from_flags():
     tf.compat.v1.enable_eager_execution()
   if FLAGS.tf_xla:
     tf.config.optimizer.set_jit(True)
-    fastmath.tf_math.set_tf_xla_forced_compile(FLAGS.tf_xla_forced_compile)
+    fastmath.tf.set_tf_xla_forced_compile(FLAGS.tf_xla_forced_compile)
   tf.config.optimizer.set_experimental_options({
       'pin_to_host_optimization': FLAGS.tf_opt_pin_to_host,
       'layout_optimizer': FLAGS.tf_opt_layout,
@@ -100,7 +98,6 @@ def _output_dir_or_default():
 
 # TODO(afrozm): Share between trainer.py and rl_trainer.py
 def _jax_and_tf_configure_for_devices():  # pylint: disable=missing-function-docstring
-  jax.config.enable_omnistaging()
   if FLAGS.use_tpu:
     jax.config.update('jax_platform_name', 'tpu')
     jax.config.update('jax_xla_backend', FLAGS.jax_xla_backend)
@@ -148,6 +145,28 @@ def tf_init_tpu(worker='', protocol=None):
     return '/job:worker'
 
 
+def _make_jax_gpu_cluster(host_id, server_ip, n_hosts, server_port=5005):
+  """Make JAX GPU Cluster."""
+
+  addr = f'{server_ip}:{server_port}'
+  if host_id == 0:
+    logging.info('starting service on %s', addr)
+    service = xc.get_distributed_runtime_service(addr, n_hosts)
+    # We add an explicit call to shutdown the service via atexit as Python
+    # interpreter may not call the service destructor on process termination.
+    atexit.register(service.shutdown)
+
+  logging.info('connecting to service on %s', addr)
+  dist_client = xc.get_distributed_runtime_client(addr, host_id)
+  dist_client.connect()
+  atexit.register(dist_client.shutdown)
+
+  # register dist gpu backend
+  factory = functools.partial(jax.lib.xla_client.make_gpu_client,
+                              dist_client, host_id)
+  jax.lib.xla_bridge.register_backend_factory('gpu', factory, priority=300)
+
+
 def main(_):
   logging.set_verbosity(FLAGS.log_level)
 
@@ -155,6 +174,13 @@ def main(_):
   _gin_parse_configs()
   _jax_and_tf_configure_for_devices()
 
+  # Create a JAX GPU cluster if using JAX and given a chief IP.
+  if fastmath.is_backend(Backend.JAX) and FLAGS.gpu_cluster_chief_ip:
+    _make_jax_gpu_cluster(FLAGS.gpu_cluster_host_id,
+                          FLAGS.gpu_cluster_chief_ip,
+                          FLAGS.gpu_cluster_n_hosts,
+                          FLAGS.gpu_cluster_port)
+
   if FLAGS.disable_jit:
     fastmath.disable_jit()
 
